from concurrent.futures import Future
from threading import RLock
import logging

_LOG = logging.getLogger('more_executors._Future')


class _Future(Future):
    # Need to reimplement some of the Future class.
    # We must hold callbacks ourselves and not let the parent class handle them.
    # This is necessary for locking: we need to handle some locking ourselves,
    # but we must NOT have the callbacks invoked while our lock is held.
    # (Note: an alternative would be to access concurrent.futures private variables)
    def __init__(self):
        super(_Future, self).__init__()
        self._me_done_callbacks = []
        self._me_lock = RLock()

    def _me_invoke_callbacks(self):
        for callback in self._me_done_callbacks:
            try:
                callback(self)
            except Exception:
                _LOG.exception('exception calling callback for %r', self)

    def add_done_callback(self, fn):
        # Overrides function from parent; intentionally does not call super,
        # so we own the callbacks ourself
        with self._me_lock:
            if not self.done():
                self._me_done_callbacks.append(fn)
                return
        # Already done -> call it directly
        fn(self)



"""Create futures which will retry implicitly.

The `more_executors.retry.RetryExecutor` class produces `Future` objects which will retry on
failure. Subclassing `more_executors.retry.RetryPolicy` allows customization of the retry behavior.
"""

from collections import namedtuple
from concurrent.futures import Executor
from threading import RLock, Thread, Event
from datetime import datetime
from datetime import timedelta

import logging

from more_executors._common import _Future

_LOG = logging.getLogger('RetryExecutor')

# Hide some docs which would otherwise be repeated
__pdoc__ = {}
__pdoc__['ExceptionRetryPolicy.should_retry'] = None
__pdoc__['ExceptionRetryPolicy.sleep_time'] = None
__pdoc__['RetryExecutor.map'] = None
__pdoc__['RetryExecutor.shutdown'] = None


def _total_seconds(td):
    # Python 2.6 compat - equivalent of timedelta.total_seconds
    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6


class RetryPolicy(object):
    """Instances of this class may be supplied to `more_executors.retry.RetryExecutor`
    to customize the retry behavior.

    This base class will never retry.  See `more_executors.retry.ExceptionRetryPolicy` for
    a general-purpose implementation."""

    def should_retry(self, attempt, result, exception):
        """Returns `True` if a `Future` should be retried.

        This method will be called after a `Future` completes or raises an exception.

        - `attempt`: number of times the future has been attempted; starts counting at 1
        - `result`: result of the future, or `None` on exception
        - `exception`: exception of the future, or `None` if no exception"""
        return False

    def sleep_time(self, attempt, result, exception):
        """Returns the amount of time (in seconds) to delay before the next
        attempt at running a `Future`.

        See `should_retry` for the meaning of each argument."""
        return 0


class ExceptionRetryPolicy(RetryPolicy):
    """Retries on any exceptions under the given base class,
    up to a fixed number of attempts, with an exponential
    backoff between each attempt."""

    def __init__(self, max_attempts, exponent, sleep, max_sleep, exception_base):
        """Create a new policy.

        - `max_attempts`: maximum number of times a `Future` should be attempted
        - `exponent`: exponent used for backoff, e.g. `2.0` will result in a delay
                      which doubles between each attempt
        - `sleep`: base value for delay between attempts in seconds, e.g. `1.0`
                   will delay by one second between first two attempts
        - `max_sleep`: maximum delay between attempts, in seconds
        - `exception_base`: `Future` will be retried if (and only if) it raises
                            an exception inheriting from this class"""
        self._max_attempts = max_attempts
        self._exponent = exponent
        self._sleep = sleep
        self._max_sleep = max_sleep
        self._exception_base = exception_base

    @classmethod
    def new_default(cls):
        """Returns a new policy with a reasonable implementation:

        - allows up to 10 attempts
        - retries on any `Exception`"""
        return cls(10, 2.0, 1.0, 120, Exception)

    def should_retry(self, attempt, return_value, exception):
        if not exception:
            return False
        if attempt >= self._max_attempts:
            return False
        return isinstance(exception, self._exception_base)

    def sleep_time(self, attempt, return_value, exception):
        return min(self._sleep * (self._exponent ** attempt), self._max_sleep)


_RetryJob = namedtuple('_RetryJob',
                       ['policy', 'delegate_future', 'future',
                        'attempt', 'when', 'fn', 'args', 'kwargs'])


class _RetryFuture(_Future):

    def __init__(self, executor):
        super(_RetryFuture, self).__init__()
        self.delegate_future = None
        self._executor = executor

    def running(self):
        with self._me_lock:
            if self.done():
                return False
            # Note that if we are not "done", but the delegate future is "done",
            # we consider ourselves as running - this should mean the delegate
            # future completed but callbacks haven't finished yet (and in fact,
            # we might decide to retry)
            if self.delegate_future:
                return self.delegate_future.running() or self.delegate_future.done()
        return False

    def _clear_delegate(self):
        with self._me_lock:
            self.delegate_future = None

    def set_result(self, result):
        with self._me_lock:
            self._clear_delegate()
            super(_RetryFuture, self).set_result(result)
        self._me_invoke_callbacks()

    def set_exception(self, exception):
        with self._me_lock:
            self._clear_delegate()
            super(_RetryFuture, self).set_exception(exception)
        self._me_invoke_callbacks()

    def cancel(self):
        with self._me_lock:
            if self.cancelled():
                return True
            if self.done():
                return False
            if not self._executor._cancel(self):
                return False
            out = super(_RetryFuture, self).cancel()
            if out:
                self.set_running_or_notify_cancel()
        if out:
            self._me_invoke_callbacks()
        return out


class RetryExecutor(Executor):
    """An `Executor` which delegates to another `Executor` while adding
    implicit retry behavior.

    - Callables are submitted to the delegate executor, and may be
      submitted more than once if retries are required.

    - The callables may be submitted to that executor from a different
      thread than the calling thread.

    - Cancelling is supported if the delegate executor allows it.

    - Cancelling between retries is always supported.

    - The returned futures from this executor are only resolved or
      failed once the callable either succeeded, or all retries
      were exhausted.  This includes activation of the done callback.

    - This executor is thread-safe.
    """

    def __init__(self, delegate, retry_policy):
        """Create a new executor.

        - `delegate`: `Executor` instance to which callables will be submitted
        - `retry_policy`: `more_executors.retry.RetryPolicy` instance used to determine
                          when futures shall be retried"""
        self._delegate = delegate
        self._default_retry_policy = retry_policy
        self._jobs = []
        self._submit_thread = Thread(
            name='RetryExecutor', target=self._submit_loop)
        self._submit_thread.daemon = True
        self._submit_event = Event()
        self._shutdown = False
        self._lock = RLock()

        self._submit_thread.start()

    def shutdown(self, wait=True):
        _LOG.info("Shutting down.")
        self._shutdown = True
        self._wake_thread()
        self._delegate.shutdown(wait)
        if wait:
            _LOG.info("Waiting for thread")
            self._submit_thread.join()
        _LOG.debug("Shutdown complete")

    @classmethod
    def new_default(cls, delegate):
        """Returns an executor with a reasonable default retry policy."""
        return cls(delegate, ExceptionRetryPolicy.new_default())

    def submit(self, fn, *args, **kwargs):
        """Submit a callable with this executor's retry policy.  Overrides `Executor.submit`.

        Returns a new `Future`.

        The callable may be invoked multiple times before the future's result
        (or exception) becomes available."""
        return self.submit_retry(
            self._default_retry_policy, fn, *args, **kwargs)

    def submit_retry(self, retry_policy, fn, *args, **kwargs):
        """Submit a callable with a specific retry policy.

        - `retry_policy`: an instance of `more_executors.retry.RetryPolicy` which
                          is used for this call only
        """
        future = _RetryFuture(self)

        job = _RetryJob(retry_policy, None, future, 0,
                        datetime.utcnow(), fn, args, kwargs)
        self._append_job(job)

        # Let the submit thread know it should wake up to check for new jobs
        self._wake_thread()

        _LOG.debug("Returning future %s", future)
        return future

    def _wake_thread(self):
        self._submit_event.set()

    def _get_next_job(self):
        # Find and return the next job to be executed, if any.
        # This means a job with when < now, or if there's none,
        # then the job with the minimum value of when.
        min_job = None
        now = datetime.utcnow()
        for job in self._jobs:
            if job.delegate_future:
                # It's already running, skip
                continue
            if job.when <= now:
                # job is overdue, just do it
                return job
            elif not min_job:
                min_job = job
            elif job.when < min_job.when:
                min_job = job
        return min_job

    def _submit_wait(self, timeout=None):
        self._submit_event.wait(timeout)
        self._submit_event.clear()

    def _submit_loop(self):
        # Runs in a separate thread continuously submitting to the delegate
        # executor until no jobs are ready, or waiting until next job is ready
        while not self._shutdown:
            _LOG.debug("_submit_loop iter")
            job = self._get_next_job()
            if not job:
                _LOG.debug("No jobs at all. Waiting...")
                self._submit_wait()
                continue

            now = datetime.utcnow()
            if job.when < now:
                # Can submit immediately and check for next job
                self._submit_now(job)
                continue

            # There is nothing to submit immediately.
            # Sleep until either:
            # - reaching the time of the nearest job, or...
            # - woken up by condvar
            delta = _total_seconds(job.when - now)
            _LOG.debug("No ready job.  Waiting: %s", delta)
            self._submit_wait(delta)

    def _submit_now(self, job):
        # Pop job since we'll replace it.
        # We need to hold the lock for the entire duration so that other
        # threads won't see _jobs between our removal and re-add of the job
        with job.future._me_lock:
            with self._lock:
                self._pop_job(job)

                # We need the future's lock now too, because someone could
                # call cancel after this check and before we submit.
                if job.future.done():
                    _LOG.debug(
                        "future done %s - not submitting to delegate",
                        job.future)
                    return

                delegate_future = self._delegate.submit(
                    job.fn, *job.args, **job.kwargs)
                job.future.delegate_future = delegate_future

                new_job = _RetryJob(
                    job.policy,
                    delegate_future,
                    job.future,
                    job.attempt + 1,
                    None,
                    job.fn, job.args, job.kwargs
                )
                self._append_job(new_job)
                _LOG.debug("Submitted: %s", new_job)

        delegate_future.add_done_callback(self._delegate_callback)
        self._wake_thread()

    def _pop_job(self, job):
        with self._lock:
            for idx, pending in enumerate(self._jobs):
                if pending is job:
                    return self._jobs.pop(idx)

    def _append_job(self, job):
        with self._lock:
            self._jobs.append(job)

    def _retry(self, job, sleep_time):
        _LOG.debug("Will retry: %s", job)

        with self._lock:
            self._pop_job(job)
            new_job = _RetryJob(
                job.policy,
                None,
                job.future,
                job.attempt,
                datetime.utcnow() + timedelta(seconds=sleep_time),
                job.fn,
                job.args,
                job.kwargs
            )
            self._append_job(new_job)

        self._wake_thread()

    def _cancel(self, future):
        found_job = None

        with self._lock:
            for idx, job in enumerate(self._jobs):
                if job.future is future:
                    _LOG.debug("Try cancel: %s", job)

                    if not job.delegate_future:
                        _LOG.debug("Successful cancel - no delegate: %s", job)
                        self._jobs.pop(idx)
                        return True

                    found_job = job
                    break

        # This shouldn't be possible.
        # - Future holds a lock on itself, and has checked that it's not already done
        # - The only other path for removing a job is in delegate_callback, but the
        #   job is only removed *after* set_result/set_exception which would wait
        #   for the future's lock.
        assert found_job, "Cancel called on orphan %s" % future

        _LOG.debug("Try cancel delegate: %s", found_job)

        if found_job.delegate_future.cancel():
            _LOG.debug("Successful cancel: %s", found_job)
            future._clear_delegate()
            # Don't remove from _jobs here,
            # the callback attached to delegate_future was expected
            # to take care of that
            return True

        _LOG.debug("Could not cancel: %s", job)
        return False

    def _delegate_callback(self, delegate_future):
        assert delegate_future.done(), \
            "BUG: callback invoked while future not done!"

        _LOG.debug("Callback activated for %s", delegate_future)

        found_job = None
        for job in self._jobs[:]:
            if job.delegate_future == delegate_future:
                found_job = job
                break

        # Callbacks are only installed after a job is added, and this is
        # the only place a job with a delegate associated will be removed,
        # thus it should not be possible for a job to be missing.
        assert found_job, ("BUG: no job associated with delegate %s" % delegate_future)

        if delegate_future.cancelled():
            # nothing to do, retrying on cancel is not allowed
            _LOG.debug("Delegate was cancelled: %s", delegate_future)
            return

        exception = delegate_future.exception()
        if exception:
            result = None
        else:
            result = delegate_future.result()

        should_retry = job.policy.should_retry(job.attempt, result, exception)
        if should_retry:
            sleep_time = job.policy.sleep_time(job.attempt, result, exception)
            self._retry(job, sleep_time)
            return

        _LOG.debug("Finalizing %s", job)

        # OK, it won't be retried.  Resolve the future.
        if exception:
            job.future.set_exception(exception)
        else:
            job.future.set_result(result)

        self._pop_job(job)

        _LOG.debug("Finalized %s", job)



from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from functools import partial

from more_executors.map import MapExecutor
from more_executors.retry import RetryExecutor, ExceptionRetryPolicy
from more_executors.poll import PollExecutor
from more_executors.cancel_on_shutdown import CancelOnShutdownExecutor


class Executors(object):
    """Convenience methods for creating executors.

    This class produces wrapped executors which may be customized
    by use of the `with_*` methods, as in the following example:

    ```
    Executors.thread_pool(max_workers=4).with_retry().with_map(lambda x: x*2)
    ```

    Produces a thread pool executor which will retry on failure
    and multiply all output values by 2."""

    _WRAP_METHODS = [
        'with_retry',
        'with_map',
        'with_poll',
        'with_cancel_on_shutdown',
    ]

    @classmethod
    def wrap(cls, executor):
        """Wrap an executor for configuration.

        This adds all `with_*` methods from this class onto the executor.
        When invoked, the executor is implicitly passed as the first argument to the method."""
        for name in cls._WRAP_METHODS:
            unbound_fn = getattr(cls, name)
            bound_fn = partial(unbound_fn, executor)
            setattr(executor, name, bound_fn)
        return executor

    @classmethod
    def thread_pool(cls, *args, **kwargs):
        """Creates a new `concurrent.futures.ThreadPoolExecutor` with the given arguments."""
        return cls.wrap(ThreadPoolExecutor(*args, **kwargs))

    @classmethod
    def process_pool(cls, *args, **kwargs):
        """Creates a new `concurrent.futures.ProcessPoolExecutor` with the given arguments."""
        return cls.wrap(ProcessPoolExecutor(*args, **kwargs))

    @classmethod
    def with_retry(cls, executor, retry_policy=None):
        """Wrap an executor in a `more_executors.retry.RetryExecutor`.

        Submitted functions will be retried on failure.

        - `retry_policy`: the `more_executors.retry.RetryPolicy` to be used; if omitted,
                          a reasonable default is applied which will retry for several minutes."""
        if not retry_policy:
            retry_policy = ExceptionRetryPolicy.new_default()
        return cls.wrap(RetryExecutor(executor, retry_policy))

    @classmethod
    def with_map(cls, executor, fn):
        """Wrap an executor in a `more_executors.map.MapExecutor`.

        Submitted callables will have their output transformed by the given function.

        - `fn`: a function used to transform each output value from the executor"""
        return cls.wrap(MapExecutor(executor, fn))

    @classmethod
    def with_poll(cls, executor, fn, cancel_fn=None, default_interval=5.0):
        """Wrap an executor in a `more_executors.poll.PollExecutor`.

        Submitted callables will have their output passed into the poll function.
        See the class documentation for more information on the poll and cancel
        functions.

        - `fn`: a function used for polling results.
        - `cancel_fn`: a function called when a future is cancelled.
        - `default_interval`: default interval between polls, in seconds."""
        return cls.wrap(PollExecutor(executor, fn, cancel_fn, default_interval))

    @classmethod
    def with_cancel_on_shutdown(cls, executor):
        """Wrap an executor in a `more_executors.cancel_on_shutdown.CancelOnShutdownExecutor`.

        Returned futures will have `cancel` invoked if the executor is shut down
        before the future has completed."""
        return cls.wrap(CancelOnShutdownExecutor(executor))



"""Create futures resolved by polling with a provided function."""
from concurrent.futures import Executor
from threading import RLock, Thread, Event
import logging

from more_executors._common import _Future

_LOG = logging.getLogger('PollExecutor')

__pdoc__ = {}
__pdoc__['PollExecutor.shutdown'] = None
__pdoc__['PollExecutor.map'] = None


class _PollFuture(_Future):
    def __init__(self, delegate, executor):
        super(_PollFuture, self).__init__()
        self._delegate = delegate
        self._executor = executor
        self._delegate.add_done_callback(self._delegate_resolved)

    def _delegate_resolved(self, delegate):
        assert delegate is self._delegate, \
            "BUG: called with %s, expected %s" % (delegate, self._delegate)

        if delegate.cancelled():
            return
        elif delegate.exception():
            self.set_exception(delegate.exception())
        else:
            self._executor._register_poll(self, self._delegate)

    def _clear_delegate(self):
        with self._me_lock:
            self._delegate = None

    def set_result(self, result):
        with self._me_lock:
            if self.done():
                return
            super(_PollFuture, self).set_result(result)
        self._me_invoke_callbacks()

    def set_exception(self, exception):
        with self._me_lock:
            if self.done():
                return
            super(_PollFuture, self).set_exception(exception)
        self._me_invoke_callbacks()

    def running(self):
        with self._me_lock:
            if self._delegate:
                return self._delegate.running()
        # If delegate is removed, we're now polling or done.
        return False

    def cancel(self):
        with self._me_lock:
            if self.cancelled():
                return True
            if self._delegate and not self._delegate.cancel():
                return False
            if not self._executor._run_cancel_fn(self):
                return False
            self._executor._deregister_poll(self)

            # Check cancelled again.
            # The reason is that we would be called recursively if
            # delegate.cancel() succeeded, and we must avoid
            # calling notify twice.
            out = super(_PollFuture, self).cancel()
            if out:
                self.set_running_or_notify_cancel()
        if out:
            self._me_invoke_callbacks()
        return out


class PollDescriptor(object):
    """A `PollDescriptor` represents an unresolved `Future`.

    The poll function used by `more_executors.poll.PollExecutor` will be
    invoked with a list of `PollDescriptor` objects.
    """

    def __init__(self, result, yield_result, yield_exception):
        """Do not construct instances of `PollDescriptor` directly."""

        self.result = result
        """The result from the delegate executor's future, which should
        be used to drive the poll."""

        self.yield_result = yield_result
        """The poll function can call this function to make the future
        yield the given result."""

        self.yield_exception = yield_exception
        """The poll function can call this function to make the future
        raise the given exception."""


class PollExecutor(Executor):
    """Instances of `PollExecutor` submit callables to a delegate `Executor`
    and resolve the returned futures via a provided poll function.

    A cancel function may also be provided to perform additional processing
    when a returned future is cancelled.

    ### **Poll function**

    The poll function has the following semantics:

    - It's called with a single argument, a list of zero or more
      `more_executors.poll.PollDescriptor` objects.

    - If the poll function can determine that a particular future should be
      completed, either successfully or in error, it should call the `yield_result`
      or `yield_exception` methods on the appropriate `PollDescriptor`.

    - If the poll function raises an exception, all futures depending on that poll
      will fail with that exception.  The poll function will be retried later.

    - If the poll function returns an int or float, it is used as the delay
      in seconds until the next poll.

    ### **Cancel function**

    The cancel function has the following semantics:

    - It's called when cancel is requested on a future returned by this executor,
      if and only if the future is currently in the list of futures being polled,
      i.e. it will not be called if the delegate future has not yet completed.

    - It's called with a single argument: the value returned by the delegate future.

    - It should return `True` if the future can be cancelled.

    - If the cancel function raises an exception, the future's cancel method will
      return `False` and a message will be logged.

    ### **Example**

    Consider a web service which executes tasks asynchronously, with an API like this:

    * To create a task: `POST https://myservice/object/:id/publish`
        * Returns an identifier for a task, such as `{"task_id": 123}`
    * To get status of a single task: `GET https://myservice/tasks/:id`
        * Returns task status, such as `{"task_id": 123, "state": "finished"}`
    * To cancel a single task: `DELETE https://myservice/tasks/:id`
    * To search for status of multiple tasks:
      `POST https://myservice/tasks/search {"task_id": [123, 456, ...]}`
        * Returns array of task statuses, such as
          `[{"task_id": 123, "state": "finished"},
            {"task_id": 456, "state": "error", "error": "..."}]`

    Imagine we want to run many tasks concurrently and monitor their status.
    Two obvious approaches include:

    * We could make a function which starts a task and polls that task until
      it completes, and have several threads run that function for different tasks.
        * This could be easily encapsulated as a future using only the executors
          from `concurrent.futures`, but is very inefficient - each task in progress
          will consume a thread for polling.
    * We could maintain a list of all outstanding tasks and implement a special-purpose
      function to wait for all of them to complete.
        * This is efficient as only one request is needed per poll interval, but since
          tasks aren't represented as futures, useful features from `concurrent.futures`
          cannot be used and different types of asynchronous work will be handled
          inconsistently.

    `PollExecutor` bridges the gap between the two approaches.  By providing a custom
    poll function, the tasks can be monitored by one thread efficiently, while at the
    same time allowing full usage of the API provided by `Future` objects.

    In this example, the poll function would look similar to this:

        def poll_tasks(poll_descriptors):
            # Collect tasks we'll need to search for
            task_ids = [d.result.get('task_id')
                        for d in poll_descriptors]

            # Search for status of these tasks
            search = {"task_id": task_ids}
            response = requests.post(
                'https://myservice/tasks/search', json=search)

            # If request failed then all outstanding futures fail
            response.raise_for_status()

            # Request passed, resolve futures accordingly
            tasks = {task.get('task_id'): task.get('state')
                    for task in response.json()}
            for d in poll_descriptors:
                task = tasks.get(d.result, {})
                state = task.get('state')
                if state == 'finished':
                    d.yield_result()
                elif state == 'error':
                    d.yield_exception(TaskFailed())

    To ensure that calling `future.cancel()` also cancels tasks on the remote service,
    we can also provide a cancel function:

        def cancel_task(task):
            task_id = task['task_id']

            # Attempt to DELETE this task
            response = requests.delete(
                'https://myservice/tasks/%s' % task_id)

            # Succeeded only if response was OK.
            # Otherwise, we may have been too late to cancel.
            return response.ok

    With the poll and cancel functions in place, futures could be created like this:

        def publish(object_id):
            response = requests.post(
                'https://myservice/object/%s/publish' % object_id)
            response.raise_for_status()
            return response.json()

        # Submit up to 4 HTTP requests at a time,
        # and poll for the returned tasks using our poll function.
        executor = Executors.\\
            threadpool(max_workers=4).\\
            with_poll(poll_tasks, cancel_task)
        futures = [executor.submit(publish, x)
                   for x in (10, 20, 30)]

        # Now we can use concurrent.futures API to get tasks as
        # the poll function determines they've completed
        for completed in as_completed(futures):
            # ...
    """

    def __init__(self, delegate, poll_fn, cancel_fn=None, default_interval=5.0):
        """Create a new executor.

        - `delegate`: `Executor` instance to which callables will be submitted
        - `poll_fn`: a polling function used to decide when futures should be resolved;
                     see the class documentation for more information
        - `cancel_fn`: a cancel function invoked when future cancel is required; see the
                       class documentation for more information
        - `default_interval`: default interval between polls (in seconds)
        """
        self._delegate = delegate
        self._default_interval = default_interval
        self._poll_fn = poll_fn
        self._cancel_fn = cancel_fn
        self._poll_descriptors = []
        self._poll_event = Event()
        self._poll_thread = Thread(name='PollExecutor', target=self._poll_loop)
        self._poll_thread.daemon = True
        self._shutdown = False
        self._lock = RLock()

        self._poll_thread.start()

    def submit(self, fn, *args, **kwargs):
        """Submit a callable.

        Returns a future which will be resolved via the poll function."""
        delegate_future = self._delegate.submit(fn, *args, **kwargs)
        out = _PollFuture(delegate_future, self)
        out.add_done_callback(self._deregister_poll)
        return out

    def _register_poll(self, future, delegate_future):
        descriptor = PollDescriptor(
            delegate_future.result(),
            future.set_result,
            future.set_exception)
        with self._lock:
            self._poll_descriptors.append((future, descriptor))
            future._clear_delegate()
            self._poll_event.set()

    def _deregister_poll(self, future):
        with self._lock:
            self._poll_descriptors = [(f, d)
                                      for (f, d) in self._poll_descriptors
                                      if f is not future]

    def _run_cancel_fn(self, future):
        if not self._cancel_fn:
            # no cancel function => no veto of cancel
            return True

        descriptor = [d
                      for (f, d) in self._poll_descriptors
                      if f is future]
        if not descriptor:
            # no record of this future => no veto of cancel.
            # we can get here if the future is already done
            # or if polling hasn't started yet
            return True

        assert len(descriptor) == 1, "Too many poll descriptors for %s" % future

        descriptor = descriptor[0]

        try:
            return self._cancel_fn(descriptor.result)
        except Exception:
            _LOG.exception("Exception during cancel on %s/%s", future, descriptor.result)
            return False

    def _run_poll_fn(self):
        with self._lock:
            descriptors = [d for (_, d) in self._poll_descriptors]
            self._poll_event.clear()

        try:
            return self._poll_fn(descriptors)
        except Exception as e:
            _LOG.debug("Poll function failed: %s", e)
            # If poll function fails, then every future
            # depending on the poll also immediately fails.
            [d.yield_exception(e) for d in descriptors]

    def _poll_loop(self):
        while not self._shutdown:
            _LOG.debug("Polling...")

            next_sleep = self._run_poll_fn()
            if not (isinstance(next_sleep, int) or isinstance(next_sleep, float)):
                next_sleep = self._default_interval

            _LOG.debug("Sleeping...")
            self._poll_event.wait(next_sleep)

    def shutdown(self, wait=True):
        self._shutdown = True
        self._poll_event.set()
        self._delegate.shutdown(wait)
        if wait:
            _LOG.debug("Join poll thread...")
            self._poll_thread.join()
            _LOG.debug("Joined poll thread.")
